/* Beholder RMON ethernet network monitor,Copyright (C) 1993 DNPAP group */
/* See file COPYING 'GNU General Public Licence' for copyright details   */
/************************************************************************
** MODULE INFORMATION **
************************
** FILE NAME:          mib.c
** SYSTEM NAME:        beholder
** MODULE NAME:        mib
** ORIGINAL AUTHOR(S): M.F.B. de Greeve
** VERSION NUMBER:     1.0
** CREATION DATE:      1992/7/13
** DESCRIPTION:        RMON Management Information Base routines
*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <agent.h>
#include <snmp.h>
#include <dnpap.h>
#include <config.h>
#include <message.h>
#include <dmath.h>
#include <dstring.h>

#include "mibsup.h"
#include "mibrqs.h"



MIB_RMON Mib =  {
                    NULL,
                    0,
                    NULL,
                    0,
                };

			  
static MIB_COMMUNITY *MibPrf = NULL;


static VOID MibExit(VOID);

static INT MibCmpProfil (MIB_COMMUNITY *Prf1, MIB_COMMUNITY *Prf2);
static INT MibCmpObject (MIB_ELEMENT *Obj1, MIB_ELEMENT *Obj2);
static INT MibCmpObjId (LONG *ObjId, WORD ObjIdLen1, LONG *ObjId2, WORD ObjIdLen2);
static WORD MibGetSetNxt (SNMP_OBJECT *Obj);



/*****************************************************************
** NAME:        MibRequest
** SYNOPSIS:    WORD MibRequest(SNMP_OBJECT *Obj)
** PARAMETERS:  Obj: pointer to MIB-object
**                  Obj->Request: MIB_PDU_GET : get-request
**                                MIB_PDU_NXT : next-request
**                                MIB_PDU_SET : set-request
**                  Obj->Id: object identification
**                  Obj->IdLen: length of Obj->Id
**                  Obj->Type: MIB_NULL: Obj->Syntax.Ptr
**                             MIB_INTEGER: Obj->Syntax.LngInt
**                             MIB_OCTETSTR: Obj->Syntax.BufChr
**                             MIB_DISPLAYSTR: Obj->Syntax.BufChr
**                             MIB_ELEMENTID: Obj->Syntax.BufInt
**                             MIB_IPADDR: Obj->Syntax.LngUns
**                             MIB_COUNTER: Obj->Syntax.LngUns
**                             MIB_GAUGE: Obj->Syntax.LngUns
**                             MIB_TIMETICKS: Obj->Syntax.LngUns
**                             MIB_OPAQUE: Obj->Syntax.BufChr
**                  Obj->Syntax: response (depends on Obj->Type)           
**                  Obj->SyntaxLen: length of Obj->Syntax            
** DESCRIPTION: searches Obj in the MIB and sets/gets its syntax.
**              MIB_PDU_SET: checks Obj->Type
**                          sets Obj->Syntax
**              MIB_PDU_GET / MIB_PDU_NXT: inserts Obj->Type
**                                       inserts Obj->Syntax
** REMARKS:     see mib.h & mibrqs.h
** RETURNS:     MIB_NOERROR : no error
**              MIB_NOSUCHNAME: Obj->Id not found in MIB or
**                              request not supported by Prf->Comm
**              MIB_BADVALUE : Obj->Request == MIB_PDU_SET and
**                             wrong Obj->Type / incorrect value
**              MIB_TOOBIG : response too big for 1 packet
**              MIB_GENERROR : general error, dependent on object
*******************************************************************/

WORD MibRequest (SNMP_OBJECT *Obj)
{
    return MibGetSetNxt(Obj);
}

/*****************************************************************
** NAME:        MibInit
** SYNOPSIS:    VOID MibInit (MIB_ELEMENT *Obj, WORD ObjSze)
** PARAMETERS:  *Obj: array of MIB-objects
**                  Obj->Id: object identification
**                  Obj->IdLen: length of Obj->Id
**                  (*Rqs)(*Obj, IdLen): function for MibGetResp
**                      *Obj: see MibGetResp
**                      IdLen: length of Obj->Id in the MIB
**                  Obj->Type: SNMP_NULL: Obj->Syntax.Ptr
**                             SNMP_INTEGER: Obj->Syntax.LngInt
**                             SNMP_OCTETSTR: Obj->Syntax.BufChr
**                             SNMP_DISPLAYSTR: Obj->Syntax.BufChr
**                             SNMP_ELEMENTID: Obj->Syntax.BufInt
**                             SNMP_IPADDR: Obj->Syntax.LngUns
**                             SNMP_COUNTER: Obj->Syntax.LngUns
**                             SNMP_GAUGE: Obj->Syntax.LngUns
**                             SNMP_TIMETICKS: Obj->Syntax.LngUns
**                             SNMP_OPAQUE: Obj->Syntax.BufChr
**                  Obj->Support: MIB_READ
**                                MIB_WRITE
**                                MIB_READ || MIB_WRITE
**              ObjSize: number of objects    
** DESCRIPTION: initializes the MIB.
**              (*Rqs) is a function-pointer to the application-
**                  specicific part of the MIB-variable. When the
**                  Obj->Id is found this function is called to
**                  retrieve the data from the application.
** REMARKS:     see also mib.h & mibsup.h
** RETURNS:     VOID
*******************************************************************/

BOOLEAN MibInit (MIB_ELEMENT *Obj, WORD ObjSze)
{				  							 
INT commnr;
CHAR configname[81];
CHAR *access;	  
AGENT_COMM *comm;
WORD support;
MIB_COMMUNITY *mibprf;
											  
	MessageConfig(MIB_ERROR, "Mib");
	
	commnr = 1;
	sprintf(configname, "mib.access.%d", commnr);
	while (ConfigGetString(configname, &access) == TRUE)
	{		     
		if ((comm = AgentCommunity(commnr)) == NULL)
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: no community %d defined in agent", commnr);
			break;
		}
		
		if (stricmp(access, "read-only") == 0)
			support = MIB_READ;
		else
		if (stricmp(access, "write-only") == 0)
			support = MIB_WRITE;
		else
		if (stricmp(access, "read-write") == 0)
			support = MIB_READ | MIB_WRITE;
		else
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: invalid access specified");
			break;
		}
		
		if ((mibprf = DnpapRealloc(MibPrf, commnr*sizeof(MIB_COMMUNITY))) == NULL)
		{
			DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: out of community name space");
			break;
		}
		
		MibPrf = mibprf;
		memcpy(MibPrf[commnr-1].Comm, comm->comm, comm->commLen);
		MibPrf[commnr-1].CommLen = comm->commLen;
		MibPrf[commnr-1].Support = support;
		
		commnr++;
		sprintf(configname, "mib.access.%d", commnr);
	}
	
	if (--commnr == 0)
	{
		DnpapMessage(DMC_ERROR, MIB_ERROR, "mib: no external access possible");
		return FALSE;
	}
	
	Mib.Prf = MibPrf;
    Mib.PrfSze = commnr;
	
    Mib.Obj = Obj;
    Mib.ObjSze = ObjSze;

    qsort (Mib.Prf, Mib.PrfSze, sizeof(MIB_COMMUNITY), MibCmpProfil);
    qsort (Mib.Obj, Mib.ObjSze, sizeof(MIB_ELEMENT), MibCmpObject);
	
	DnpapAtExit(MibExit);

    return TRUE;
}


VOID MibExit(VOID)
{
	DnpapFree(MibPrf);
}
 

/*****************************************************************
** NAME:        MibSimple
** SYNOPSIS:    BOOLEAN MibSimple (SNMP_OBJECT *Obj,unsigned IdLen)
** PARAMETERS:  *Obj: see MibRequest()
**              IdLen: length of Obj->Id in the MIB
** DESCRIPTION: checks the Obj->Id of a simple object.
** REMARKS:     see also MibRmon(), mib.h & mibsup.h
** RETURNS:     TRUE: good Obj->Id: process further
**              FALSE: bad Obj->Id: return SNMP_NOSUCHNAME
*******************************************************************/

BOOLEAN MibSimple (SNMP_OBJECT *Obj, WORD IdLen)
{
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen == IdLen + 1 && Obj->Id[IdLen] == 0)
            return TRUE;
        return FALSE;
    }
    if (Obj->IdLen == IdLen)
    {
        Obj->Id[IdLen] = 0;
        Obj->IdLen++;
        return TRUE;
    }
    return FALSE;
}

/*****************************************************************
** NAME:        MibRmon
** SYNOPSIS:    MIB_LOCAL * MibRmon (SNMP_OBJECT *Obj,
**                  MIB_LOCAL *Local, WORD IdLen, WORD IdSize)
** PARAMETERS:  *Obj: see MibRequest()
**              *Local: begin of list of local datastructures
**              IdLen: length of Obj->Id in the MIB
**              IdSize: length of Obj->Id after the MIB
** DESCRIPTION: checks the Obj->Id of an Rmon (aggregate) object.
** REMARKS:     see also MibSimple(), mib.h & mibsup.h
** RETURNS:     pointer to found local datastructure.
**                  ->Index: collector index
**                  ->Data: pointer to the application data
**              NULL: local datastructure was not found
*******************************************************************/

MIB_LOCAL * MibRmon (SNMP_OBJECT *Obj, MIB_LOCAL *Local, WORD IdLen, WORD IdSize)
{
#ifndef MARC
    if (Local == NULL)
        return NULL;
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen != IdLen + IdSize)
			return NULL;
        while (Local != NULL && Local->Index < Obj->Id[IdLen])
            Local = Local->Next;
		if (Local == NULL || Local->Index != Obj->Id[IdLen])
            return NULL;
		return Local;
    }
    if (Obj->IdLen == IdLen)        /* first instance */
        return Local;
    while (Local != NULL && Local->Index < Obj->Id[IdLen])
        Local = Local->Next;
	if (Local == NULL || Local->Index < Obj->Id[IdLen])
        return NULL;
    return Local;
#else
    if (Local == NULL)
        return NULL;
    if (Obj->Request != SNMP_PDU_NEXT)
    {
        if (Obj->IdLen != IdLen + IdSize)
            return NULL;
        while (Local->Index < Obj->Id[IdLen] && Local->Next != NULL)
            Local = Local->Next;
        if (Local->Index == Obj->Id[IdLen])
            return Local;
        return NULL;
    }
    if (Obj->IdLen > IdLen + IdSize)        /* bad index */
        return NULL;
    if (Obj->IdLen == IdLen)                /* first instance */
        return Local;
    while (Local->Index < Obj->Id[IdLen] && Local->Next != NULL)
        Local = Local->Next;
    if (Local->Index >= Obj->Id[IdLen])
        return Local;
    return NULL;
#endif
}

/*****************************************************************
** NAME:        MibInsert
** SYNOPSIS:    MIB_LOCAL * MibInsert (SNMP_OBJECT *Obj,
**                  MIB_LOCAL **local, WORD IdLen, WORD IdSize)
** PARAMETERS:  *Obj: see MibRequest()
**              **local: begin of list of local datastructures
**              IdLen: length of Obj->Id in the MIB
**              IdSize: length of Obj->Id after the MIB
** DESCRIPTION: creates a new Rmon instance in local datastructure.
** REMARKS:     see also MibRemove(), mib.h & mibsup.h
** RETURNS:     pointer to created instance.
**                  Index: collector index
**                  Data: pointer to the application data
**              NULL: local datastructure was not created
*******************************************************************/

MIB_LOCAL * MibInsert
    (SNMP_OBJECT *Obj, MIB_LOCAL **local, WORD IdLen, WORD IdSize)
{
    MIB_LOCAL * Junk;
    MIB_LOCAL * Local = *local;

    if (Obj->IdLen != IdLen + IdSize || Obj->Id[IdLen] < 1)
        return NULL;
    if ((Junk = DnpapMalloc (sizeof(MIB_LOCAL))) == NULL)
        return NULL;
    Junk->Index = Obj->Id[IdLen];
    
    if (Local == NULL || Local->Index > Junk->Index)
    {
        Junk->Next = Local;
        *local = Junk;
        return Junk;
    }
    while (Local->Next != NULL && Local->Next->Index <= Junk->Index)
        Local = Local->Next;
    if (Local->Index == Junk->Index)
    {
        DnpapFree(Junk);
        return Local;
    }
    Junk->Next = Local->Next;
    Local->Next = Junk;

    return Junk;
}

/*****************************************************************
** NAME:        MibRemove
** SYNOPSIS:    BOOLEAN MibRemove (SNMP_OBJECT *Obj,
**                  MIB_LOCAL **Local, WORD IdLen, WORD IdSize)
** PARAMETERS:  *Obj: see MibRequest()
**              **Local: begin of list of Local datastructures
**              IdLen: length of Obj->Id in the MIB
**              IdSize: length of Obj->Id after the MIB
** DESCRIPTION: removes an Rmon instance in Local datastructure.
** REMARKS:     see also MibInsert(), mib.h & mibsup.h
** RETURNS:     TRUE: instance was removed
**              FALSE: instance was not found
*******************************************************************/

BOOLEAN MibRemove (SNMP_OBJECT *Obj, MIB_LOCAL **local, WORD IdLen, WORD IdSize)
{
    MIB_LOCAL * Junk;
    MIB_LOCAL * Local = *local;

    if (Local == NULL)
        return FALSE;
    if (Obj->IdLen != IdLen + IdSize)
        return FALSE;
    if (Local->Index == Obj->Id[IdLen])
    {
        Junk = Local->Next;
        DnpapFree(Local);
        *local = Junk;
        return TRUE;
    }
    while (Local->Next != NULL && Local->Next->Index < Obj->Id[IdLen])
        Local = Local->Next;
    if (Local->Next->Index == Obj->Id[IdLen])
    {
        Junk = Local->Next->Next;
        DnpapFree(Local->Next);
        Local->Next = Junk;
        return TRUE;
    }
    return FALSE;
}

INT MibCmpProfil (MIB_COMMUNITY *Prf1, MIB_COMMUNITY *Prf2)
{
    return memcmp(Prf1->Comm, Prf2->Comm, SNMP_SIZE_COMM); 
}

INT MibCmpObject (MIB_ELEMENT *Obj1, MIB_ELEMENT *Obj2)
{
    return MibCmpObjId (Obj1->Id, Obj1->IdLen, Obj2->Id, Obj2->IdLen);
}

INT MibCmpObjId (LONG *ObjId1, WORD ObjIdLen1, LONG *ObjId2, WORD ObjIdLen2)
{
    while (ObjIdLen1 > 0 && ObjIdLen2 > 0 && *ObjId1 == *ObjId2)
    {
        ObjIdLen1--;
        ObjIdLen2--;
        ObjId1++;
        ObjId2++;
    }
    if (!ObjIdLen1 && !ObjIdLen2)
        return 0;
    if (!ObjIdLen1)
        return -1;
    if (!ObjIdLen2)
        return 1;
    if (*ObjId1 < *ObjId2)
        return -2;
    else
        return 2;
}

WORD MibProfil (BYTE *Community, WORD CommunityLen)
{
    INT Fst, Lst, Ind;

    Fst = 0;
    Lst = Mib.PrfSze-1;

    while (Fst < Lst)
    {
        Ind = (Fst + Lst)/2;
        if (memcmp(Mib.Prf[Ind].Comm, Community, Mib.Prf[Ind].CommLen) < 0)
            Fst = Ind + 1;
        else
            Lst = Ind;
    }
    while (Fst < Mib.PrfSze && !memcmp(Mib.Prf[Fst].Comm, Community, Mib.Prf[Fst].CommLen))
    {
        if (Mib.Prf[Fst].CommLen == CommunityLen)
            return Mib.Prf[Fst].Support;
        Fst++;
    }
    return 0;
}

WORD MibGetSetNxt (SNMP_OBJECT *Obj)
{
    INT Err, Fst, Lst, Ind;
	SNMP_OBJECT SaveObj;

    Err = SNMP_NOSUCHNAME;
    Fst = 0;
    Lst = Mib.ObjSze-1;

    while (Fst < Lst)
    {
        Ind = (Fst + Lst)/2;
        switch (MibCmpObjId(Mib.Obj[Ind].Id, Mib.Obj[Ind].IdLen, 
						    Obj->Id, Obj->IdLen))
        {
        case -2:
            Fst = Ind+1;
            break;
        case -1:
        case 0:
            Fst = Lst = Ind;
            break;
        case 1:
        case 2:
            Lst = Ind;
            break;
        }
    }
    switch (Obj->Request)
    {
        case SNMP_PDU_GET:
            if (!(Mib.Obj[Fst].Support & MIB_READ))
                return Err;
            if (Mib.Obj[Fst].Rqs == NULL)
                return Err;
            if (!(MibCmpObjId (Mib.Obj[Fst].Id, Mib.Obj[Fst].IdLen, Obj->Id, Obj->IdLen)==-1))
				return Err;
            Err = Mib.Obj[Fst].Rqs (Obj, Mib.Obj[Fst].IdLen);
            if (Err == SNMP_NOERROR)
                Obj->Type = Mib.Obj[Fst].Type;
            return Err;
        case SNMP_PDU_SET:
            if (!(Mib.Obj[Fst].Support & MIB_WRITE))
                return Err;
            if (Mib.Obj[Fst].Rqs == NULL)
                return Err;
            if (!(MibCmpObjId (Mib.Obj[Fst].Id, Mib.Obj[Fst].IdLen, Obj->Id, Obj->IdLen)==-1))
				return Err;
            if (Mib.Obj[Fst].Type != Obj->Type)
                return SNMP_BADVALUE;
            if (Obj->Type == SNMP_OBJECTID && Obj->SyntaxLen > SNMP_SIZE_BUFINT)
                return SNMP_BADVALUE;
            if ((Obj->Type == SNMP_OCTETSTR ||
                 Obj->Type == SNMP_DISPLAYSTR ||
                 Obj->Type == SNMP_OPAQUE) && 
				(Obj->SyntaxLen > SNMP_SIZE_BUFCHR))
            	return SNMP_BADVALUE;
            Err = Mib.Obj[Fst].Rqs (Obj, Mib.Obj[Fst].IdLen);
            return Err;
        case SNMP_PDU_NEXT:
            switch (MibCmpObjId(Mib.Obj[Fst].Id, Mib.Obj[Fst].IdLen, 
								Obj->Id, Obj->IdLen))
            {
            case -2:
                return Err;
            case -1:
            case 0:
                if (Mib.Obj[Fst].Rqs != NULL && Mib.Obj[Fst].Support & MIB_READ)
                {
                    Err = Mib.Obj[Fst].Rqs (Obj, Mib.Obj[Fst].IdLen);
                    if (Err == SNMP_NOERROR)
                        Obj->Type = Mib.Obj[Fst].Type;
                }
                Fst++;
				break;
            }
            if (Err == SNMP_NOSUCHNAME)
				memcpy(SaveObj.Id, Obj->Id, 
					   (SaveObj.IdLen = Obj->IdLen) * sizeof(Obj->Id[0]));
            while (Fst < Mib.ObjSze && Err == SNMP_NOSUCHNAME)
            {
                memcpy(Obj->Id, Mib.Obj[Fst].Id,
                       Mib.Obj[Fst].IdLen * sizeof(LONG));
                Obj->IdLen = Mib.Obj[Fst].IdLen;
                if (Mib.Obj[Fst].Rqs != NULL && Mib.Obj[Fst].Support & MIB_READ)
                {
                    Err = Mib.Obj[Fst].Rqs (Obj, Mib.Obj[Fst].IdLen);
                    if (Err == SNMP_NOERROR)
                        Obj->Type = Mib.Obj[Fst].Type;
                }
                Fst++;
            }
            if (Err == SNMP_NOSUCHNAME)
				memcpy(Obj->Id, SaveObj.Id, 
					   (Obj->IdLen = SaveObj.IdLen) * sizeof(Obj->Id[0]));
            return Err;
    }
	
	return SNMP_GENERROR;
}
